﻿using System;
using System.Linq;
using System.Text;

namespace TestCoverageAnalysisTool
{
  public class Matrix : IEquatable<Matrix>
  {
    private readonly int _dimension;

    private int[][] _fields;

    public Matrix(int dimension)
    {
      _dimension = dimension;

      _fields = new int[_dimension][];

      for (int i = 0; i < _dimension; i++)
      {
        _fields[i] = new int[_dimension];
      }
    }

    public Matrix(Matrix source)
      : this(source.ToArray())
    {

    }

    public Int32 Dimension
    {
      get { return _dimension; }
    }

    public Matrix(int[,] array)
    {
      _dimension = array.GetLength(0);

      _fields = new int[_dimension][];

      for (int i = 0; i < _dimension; i++)
      {
        _fields[i] = new int[_dimension];
      }

      for (int i = 0; i < array.GetLength(0); i++)
      {
        for (int j = 0; j < array.GetLength(1); j++)
        {
          _fields[i][j] = array[i, j];
        }
      }
    }

    public int this[int index1, int index2]
    {
      get
      {
        return _fields[index1][index2];
      }
      set
      {
        _fields[index1][index2] = value;
      }
    }

    public int[] Row(int index)
    {
      return _fields[index];
    }

    public int[] Column(int index)
    {
      var result = new int[_dimension];

      for (int i = 0; i < _dimension; i++)
      {
        result[i] = _fields[i][index];
      }

      return result;
    }

    public Boolean IsRowZero(int index)
    {
      for (int i = 0; i < _dimension; i++)
      {
        if (_fields[index][i] == 1)
        {
          return false;
        }
      }

      return true;
    }

    //Замена строки
    public void SetRow(int index, int[] array)
    {
      for (int i = 0; i < array.Length; i++)
      {
        _fields[index][i] = array[i];
      }
    }

    //Замена столбца
    public void SetColumn(int index, int[] array)
    {
      for (int i = 0; i < array.Length; i++)
      {
        _fields[i][index] = array[i];
      }
    }

    //Перевод матрицы в двумерный массив
    public int[,] ToArray()
    {
      var result = new int[_dimension, _dimension];

      for (int i = 0; i < _dimension; i++)
      {
        for (int j = 0; j < _dimension; j++)
        {
          result[i, j] = _fields[i][j];
        }
      }

      return result;
    }

    //Равенство матриц
    public bool Equals(Matrix other)
    {
      for (int i = 0; i < _dimension; i++)
      {
        for (int j = 0; j < _dimension; j++)
        {
          if (this[i, j] != other[i, j])
            return false;
        }
      }

      return true;
    }

    //Транспонирование матрицы
    public Matrix Transpose()
    {
      var result = new Matrix(_dimension);

      for (int i = 0; i < _dimension; i++)
      {
        for (int j = 0; j < _dimension; j++)
        {
          result[j, i] = this[i, j];
        }
      }

      return result;
    }

    //Симметричное отображение относительно горизонтальной оси
    public Matrix HorizontalMapping()
    {
      var result = new Matrix(_dimension);

      for (int i = 0; i <= _dimension / 2; i++)
      {
        result.SetRow(i, this.Row(_dimension - i - 1));
        result.SetRow(_dimension - i - 1, this.Row(i));
      }

      return result;
    }

    //Симметричное отображение относительно вертикальной оси
    public Matrix VerticalMapping()
    {
      var result = new Matrix(_dimension);

      for (int i = 0; i <= _dimension / 2; i++)
      {
        result.SetColumn(i, this.Column(_dimension - i - 1));
        result.SetColumn(_dimension - i - 1, this.Column(i));
      }

      return result;
    }

    public string Vis
    {
      get
      {
        var strBuilder = new StringBuilder();

        for (int i = 0; i < _dimension; i++)
        {
          strBuilder.Append("{");

          for (int j = 0; j < _dimension; j++)
          {
            strBuilder.Append(_fields[i][j] + ((j == _dimension - 1) ? "" : " "));
          }

          strBuilder.Append("}");
          strBuilder.AppendLine();
        }

        return strBuilder.ToString();
      }

    }

    public override string ToString()
    {
      var strBuilder = new StringBuilder();

      for (int i = 0; i < _dimension; i++)
      {
        strBuilder.Append("{");

        for (int j = 0; j < _dimension; j++)
        {
          strBuilder.Append(_fields[i][j] + ((j == _dimension - 1) ? "" : " "));
        }

        strBuilder.Append("} ");
      }

      return strBuilder.ToString();
    }

    //Поэлементное сложение матриц
    public static Matrix MatrixAddition(Matrix mtr1, Matrix mtr2)
    {
      int dimension = mtr1.Dimension;
      var result = new Matrix(dimension);

      for (int i = 0; i < dimension; i++)
      {
        for (int j = 0; j < dimension; j++)
        {
          result[i, j] = mtr1[i, j] + mtr2[i, j];
        }
      }

      return result;
    }

    //Умножение матриц
    public static Matrix MatrixMultiplication(Matrix mtr1, Matrix mtr2)
    {
      int dimension = mtr1.Dimension;
      var result = new Matrix(dimension);

      for (int i = 0; i < dimension; i++)
      {
        for (int j = 0; j < dimension; j++)
        {
          result[i, j] = DotProduct(mtr1.Row(i), mtr2.Column(j));
        }
      }

      return result;
    }

    //Скалярное произведение векторов
    public static int DotProduct(int[] vector1, int[] vector2)
    {
      int result = 0;

      for (int i = 0; i < vector1.Count(); i++)
      {
        result += vector1[i] * vector2[i];
      }

      return result;
    }

    //Функция Sign от матрицы 
    public static Matrix Sign(Matrix mtr)
    {
      int dimension = mtr.Dimension;
      var result = new Matrix(dimension);

      for (int i = 0; i < dimension; i++)
      {
        for (int j = 0; j < dimension; j++)
        {
          result[i, j] = mtr[i, j] > 0 ? 1 : 0;
        }
      }
      return result;
    }

    public static Matrix GetIdentityMatrix(int dimension)
    {
      var initMatrix = new Matrix(dimension);

      for (int i = 0; i < dimension; i++)
      {
        for (int j = 0; j < dimension; j++)
        {
          if (i == j)
          {
            initMatrix[i, j] = 1;
          }
          else
          {
            initMatrix[i, j] = 0;
          }
        }
      }
      return initMatrix;
    }

    //Алгоритм Флойда-Уоршелла для получения матрицы достижимости по матрице смежности
    public static Matrix GetReachabilityMatrix(Matrix adjMtr)
    {
      int dimension = adjMtr.Dimension;
      var result = new Matrix(dimension);

      var currentTime = System.DateTime.Now;

      for (int k = 0; k < dimension; k++)
      {
        for (int i = 0; i < dimension; i++)
        {
          for (int j = 0; j < dimension; j++)
          {
            result[i, j] = adjMtr[i, j] | (adjMtr[i, k] & adjMtr[k, j]);
            Console.WriteLine("({0}, {1}, {2})", k, i, j);
          }
        }
      }

      var elapsedTime = System.DateTime.Now - currentTime;

      Console.WriteLine("Матрица достижимости получена");

      return result;
    }

    //Получение матрицы сильной связности по матрице достижимости
    public static Matrix GetStrongConnectivityMatrix(Matrix reachabilityMatrix)
    {
      var dimension = reachabilityMatrix.Dimension;
      var result = new Matrix(dimension);

      for (int i = 0; i < dimension; i++)
      {
        for (int j = 0; j < dimension; j++)
        {
          result[i, j] = reachabilityMatrix[i, j] & reachabilityMatrix[j, i];
        }
      }

      Console.WriteLine("Матрица сильной связности получена");

      return result;
    }
  }
}